from zenlog import log
from dataclasses import dataclass, field
from mccode_antlr.instr import Instr
from mccode_antlr.common import Expr
from nexusformat.nexus import NXfield, NXcollection


@dataclass
class NXInstr:
    instr: Instr
    declared: dict[str, Expr] = field(default_factory=dict)
    nxlog_root: str = field(default_factory=str)

    def __post_init__(self):
        """Start the C translation to ensure McCode-oddities are handled before any C-code parsing."""
        from mccode_antlr.common import ShapeType, DataType, Value
        from mccode_antlr.translators.target import MCSTAS_GENERATOR
        from mccode_antlr.translators.c import CTargetVisitor, CDeclaration
        from mccode_antlr.translators.c_listener import evaluate_c_defined_expressions
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False, output=None)
        translator = CTargetVisitor(self.instr, generate=MCSTAS_GENERATOR, config=config)
        # translator.instrument_uservars is a list of `CDeclaration` objects, which are named tuples with
        # fields: name type init is_pointer is_array orig
        # translator.component_uservars is a dictionary of lists for each component type of `CDeclaration` objects.

        # only worry about instrument level variables for the moment, and convert the CDeclarations into Expr objects
        def c_declaration_to_expr(dec: CDeclaration) -> Expr:
            expr = Expr(Value(None)) if dec.init is None else Expr.parse(dec.init)
            expr.data_type = DataType.from_name(dec.type)
            if dec.is_pointer or dec.is_array:
                expr.shape_type = ShapeType.vector
            return expr

        variables = {dec.name: c_declaration_to_expr(dec) for dec in translator.instrument_uservars}

        # defined as
        # TODO this does not work because the simple "C"-style expression parser doesn't know about pointers
        # Hopefully any %include style lines have been removed at this point.
        all_inits = '\n'.join(init.source for init in self.instr.initialize)
        try:
            variables = evaluate_c_defined_expressions(variables, all_inits)
        except AttributeError:
            log.warn(f'Evaluating INITIALIZE %{{\n{all_inits}%}}\n failed; see preceding errors for hints why. '
                     'This is not an error condition (for now). Continuing')

        self.declared = variables

    def to_nx(self):
        # quick and very dirty:
        return NXfield(str(self.instr))

    def expr2nx(self, expr: Expr):
        from moreniius.utils import link_specifier
        if isinstance(expr, list):
            return [self.expr2nx(x) for x in expr]
        if isinstance(expr, tuple):
            return tuple([self.expr2nx(x) for x in expr])
        if not isinstance(expr, Expr):
            return expr

        if expr.is_constant:
            return expr.value

        evaluated = expr.evaluate(self.declared)
        if evaluated.is_constant:
            return evaluated.value

        dependencies = [par.name for par in self.instr.parameters if evaluated.depends_on(par.name)]
        if len(dependencies):
            log.warn(f'The expression {expr} depends on instrument parameter(s) {dependencies}\n'
                     f'A link will be inserted for each; make sure their values are stored at {self.nxlog_root}/')
            links = {par: link_specifier(par, f'{self.nxlog_root}/{par}') for par in dependencies}
            return NXcollection(expression=str(expr), **links)

        return str(expr)

    def make_nx(self, nx_class, *args, **kwargs):
        nx_args = [self.expr2nx(expr) for expr in args]
        nx_kwargs = {name: self.expr2nx(expr) for name, expr in kwargs.items()}
        return nx_class(*nx_args, **nx_kwargs)



from mccode_antlr.instr import Instance
from nexusformat.nexus import NXevent_data

class NotNXdict:
    """Wrapper class to prevent NXfield-parsing of the held dictionary"""
    def __init__(self, v: dict):
        self.value = v

    def to_json_dict(self):
        return self.value

    def __repr__(self):
        return f"NotNXdict<{self.value}>"


def outer_transform_dependency(transformations):
    """For a NXtransformations group, find the most-dependent transformation name

    E.g., for
    transforms:NXtransformations
      rotation_angle
        @depends_on=.
      chi
        @depends_on=rotation_angle
      phi
        @depends_on=chi

    find and return 'phi' since it depends on 'chi', which depends on 'rotation_angle', which is independent

    The dependency chain *must* be singular and fully contained in the NXtransformations object for this to work
    """
    names = list(transformations)
    if len(names) == 1:
        return names[0]
    depends = {name: getattr(transformations, name).depends_on for name in names}
    externals = [v for k, v in depends.items() if v not in depends]
    if len(externals) != 1:
        raise RuntimeError(f"Dependency chain {depends} should have one absolute dependency, found {externals} instead")

    def dep_of(name):
        d = [k for k, v in depends.items() if v == name]
        if len(d) != 1:
            raise RuntimeError(f'Expected one dependency of {name} but found dependencies: {d}')
        return d[0]

    chain = [dep_of(externals[0])]
    while len(chain) < len(names):
        chain.append(dep_of(chain[-1]))
    return chain[-1]


def _sanitize(indict):
    for k in indict:
        if isinstance(indict[k], str):
            indict[k] = indict[k].replace('<nl>', '\n').replace('<tb>', '    ')\
                .replace('<qt>', "'").replace('<bs>', '\\')
        elif isinstance(indict[k], dict):
            indict[k] = _sanitize(indict[k])
    return indict


def dict2NXobj(indict, only_nx=True):
    from nexusformat.nexus import NXfield
    import nexusformat.nexus as nexus
    outdict = {}
    def type_is_ok(t):
        return not only_nx or t.startswith('NX')

    for k, v in indict.items():
        if isinstance(v, dict) and all([f in v for f in ['type', 'value']]) and type_is_ok(v['type']):
            if v['type'] == 'NXfield':
                outdict[k] = NXfield(v['value'], **v.pop('attributes', {}))
            elif not only_nx and v['type'] == 'dict':
                outdict[k] = NotNXdict(v['value'])
            else:
                nxobj = getattr(nexus, v.pop('type'))
                outdict[k] = nxobj(**v.pop('value'))
        else:
            outdict[k] = v
    return outdict


def get_mcstasscript_component_eniius_data(comp):
    """Checks for the post-McStas 3.3 METADATA attribute, and looks there for JSON eniius_data.
    Always checks EXTEND blocks as well, for `char eniius_data[]` style entries
    """
    from re import compile

    ed = 'eniius_data'
    # METADATA is a dict[str, (str, str)] where the keys are 'names' and the tuple is (type, value)
    if hasattr(comp, 'METADATA') and ed in comp.METADATA and comp.METADATA[ed][0].lower() == 'json':
        return comp.METADATA[ed][1]
    # match any '{name}eniius_data{suffix}={encoded data};
    ed_regex = compile(rf'.*{ed}[^=]*=([^;]*);')
    ed_match = ed_regex.match(comp.EXTEND)
    if not ed_match:
        return None
    to_translate = ed_match.group(1)
    # extract the matched group, remove all \n, \, and "; replace all ' by "
    json_str = to_translate.translate(str.maketrans("'", '"', '\n"\\'))
    return json_str


def decode_component_eniius_data(comp, only_nx=True) -> dict:
    """Extract the 'eniius_data' block from the component EXTEND statement

    If found, decode the information and return a dictionary of contained NXobjects
    """
    from json import JSONDecodeError, loads
    json_str = get_mcstasscript_component_eniius_data(comp)
    if json_str is None:
        return {}
    try:
        return dict2NXobj(_sanitize(loads(json_str)), only_nx=only_nx)
    except SyntaxError:
        return {}
    except JSONDecodeError as er:
        print(f"failed to decode\n{json_str}\ndue to error {er}")
        return {}


def mccode_component_eniius_data(comp: Instance, only_nx=True) -> dict:
    """Require that any eniius data comes from a METADATA tag"""
    from json import JSONDecodeError, loads
    json_str = [md.value for md in comp.metadata if md.name == 'eniius_data' and md.mimetype.lower() == 'json']
    if not json_str:
        return {}
    if len(json_str) > 1:
        print(f'{len(json_str)} "eniius_data" METADATA blocks, using only the first one')
    try:
        return dict2NXobj(loads(json_str[0]), only_nx=only_nx)
    except JSONDecodeError as er:
        print(f'Failed to decode {json_str[0]} due to error {er}')
        return {}


def ess_flatbuffer_specifier(module: str, config: dict) -> NotNXdict:
    """ESS uses flat buffers to communicate NeXus file content information, this packages one for eniius processing"""
    return NotNXdict({'module': module, 'config': config})


def ev44_event_data_group(source: str, topic: str) -> NXevent_data:
    """Produce an NXevent_data group with ev44 stream information specified to insert group-elements

    Parameters:
        source: str
            the name of the event producer, typically matches an instrument detector name
        topic: str
            the name of the Kafka stream on which the to-be-read events are published
    """
    return NXevent_data(data=ess_flatbuffer_specifier('ev44', {'source': source, 'topic': topic}))


def link_specifier(name: str, source: str) -> NotNXdict:
    """
    Constructs a specifier to insert a NeXus link into an ESS produced NeXus file

    Parameters:
        name: str
            the name given to the NeXus link, should be unique at its group level
        source: str
            the target of the link, the location in the eventual NeXus file where the source of data resides

    Examples:
    To produce the link in the following NeXus JSON structure, one would use this function
        {
          "name": "data",
          "type": "group",
          "attributes": [
            {
              "name": "NX_class",
              "dtype": "string",
              "values": "NXdata"
            }
          ],
          "children": [
            {
              "module": "link",
              "config": {
                "name": "detector0_event_data",
                "source": "/entry/instrument/detector_panel_0/event_data"
              }
            },
          ]
        }
    one would use this function to produce an NotNXdict that the eniius writer will insert into the resulting JSON
    file _without_ converting its contents to NX* objects first.

    >>> link_specifier('detector0_event_data', '/entry/instrument/detector_panel_0/event_data')
    """
    return ess_flatbuffer_specifier('link', {'name': name, 'source': source})
